/*-
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1988, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
#if 0
static char sccsid[] = "@(#)fstat.c	8.3 (Berkeley) 5/2/95";
#endif
static const char rcsid[] =
  "$FreeBSD: src/usr.bin/fstat/fstat.c,v 1.16.2.2 1999/08/29 15:28:06 peter Exp $";
#endif /* not lint */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/stat.h>
#include <sys/vnode.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/domain.h>
#include <sys/protosw.h>
#include <sys/un.h>
#include <sys/unpcb.h>
#include <sys/sysctl.h>
#include <sys/filedesc.h>
#include <sys/queue.h>
#include <sys/pipe.h>
#include <sys/mbuf.h>
#define	_KERNEL
#include <sys/file.h>
#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>
#undef _KERNEL
#include <sys/mount.h>
#include <nfs/nfsproto.h>
#include <nfs/rpcv2.h>
#include <nfs/nfs.h>
#include <nfs/nfsnode.h>

#include <vm/vm.h>
#include <vm/vm_param.h>
#if __FreeBSD__ == 3
#include <vm/vm_prot.h>
#endif
#include <sys/lock.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>

#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>

#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <kvm.h>
#include <limits.h>
#include <nlist.h>
#include <paths.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>

struct nlist nl[] = {
#define NL_MB_MAP 0
	{ "_mb_map" },
#define NL_MCLREFCNT 1
	{ "_mclrefcnt" },
#define NL_MMBFREE 2
	{ "_mmbfree" },
#define NL_MCLFREE 3
	{ "_mclfree" },
	{ NULL },
};

int 	fsflg,	/* show files on same filesystem as file(s) argument */
	pflg,	/* show files open by a particular pid */
	uflg;	/* show files open by a particular (effective) user */
int 	checkfile; /* true if restricting to particular files or filesystems */
int	nflg;	/* (numerical) display f.s. and rdev as dev_t */
int	vflg;	/* display errors in locating kernel data objects etc... */

/*
 * a kvm_read that returns true if everything is read
 */
#define KVM_READ(kaddr, paddr, len) \
	(kvm_read(kd, (u_long)(kaddr), (char *)(paddr), (len)) == (len))

kvm_t *kd;

void usage __P((void));

void do_mbuf(int i, struct mbuf *m, struct mbuf *m_kern);
void dump_mbuf(int i, struct mbuf *m, struct mbuf *m_kern);

char *mclrefcnt, *mclrefcnt_kern;
int *minfo;
int *mprevi;
struct mbuf *mbuf;
struct mbuf *mbutl;
int dumpflag = 0;

#define MINFO_VALIDMBUF	0x0001
#define MINFO_FREEMBUF	0x0002
#define MINFO_SEEN	0x0004

int
main(argc, argv)
	int argc;
	char **argv;
{
	int arg, ch, what;
	char *memf, *nlistf;
	char buf[_POSIX2_LINE_MAX];
	struct vm_map mb_map;
	caddr_t mb_map_off;
	int size, allocated, nmbufs, i, nmcl, n, index;
	void *vp, *mclfree, *mmbfree;

	arg = 0;
	what = KERN_PROC_ALL;
	nlistf = memf = NULL;
	while ((ch = getopt(argc, argv, "N:M:xX")) != -1) {
		switch((char)ch) {
		case 'M':
			memf = optarg;
			break;
		case 'N':
			nlistf = optarg;
			break;
		case 'x':
			dumpflag = 1;
			break;
		case 'X':
			dumpflag = 2;
			break;
		case '?':
		default:
			usage();
		}
	}

	/*
	 * Discard setgid privileges if not the running kernel so that bad
	 * guys can't print interesting stuff from kernel memory.
	 */
	if (nlistf != NULL || memf != NULL)
		setgid(getgid());

	if ((kd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, buf)) == NULL)
		errx(1, "%s", buf);
	if (kvm_nlist(kd, nl) != 0)
		errx(1, "no namelist: %s", kvm_geterr(kd));

	if (!KVM_READ(nl[NL_MB_MAP].n_value, &mb_map_off, sizeof(mb_map_off)))
		errx(1, "kvm_read: %s", kvm_geterr(kd));

	if (!KVM_READ(mb_map_off, &mb_map, sizeof(mb_map)))
		errx(1, "kvm_read: %s", kvm_geterr(kd));

	size =  mb_map.header.end - mb_map.header.start;
	allocated = mb_map.size;
	printf("mb_map: %d/%d K used (%.2f%%)\n", allocated >> 10, size >> 10,
	    (double)allocated/(double)size*100.0);
	mbutl = (struct mbuf *)mb_map.header.start;
	nmbufs = allocated/MSIZE;
	nmcl = allocated/MCLBYTES;
	printf("map start: %p, end %p\n %d/%d pages used, %d mbuf-size slots\n",
	    mbutl, (caddr_t)mb_map.header.end, allocated/4096, size/4096,
	    nmbufs);
	
	mclrefcnt = malloc(nmcl);
	if (!KVM_READ(nl[NL_MCLREFCNT].n_value, &mclrefcnt_kern,
	    sizeof(mclrefcnt_kern)))
		errx(1, "kvm_read: %s", kvm_geterr(kd));
	if (!KVM_READ(mclrefcnt_kern, mclrefcnt, nmcl))
		errx(1, "kvm_read: %s", kvm_geterr(kd));

	printf("Checking cluster ref counts\n");
	for (i = 0; i < nmcl; i++) {
		if (mclrefcnt[i] < 0) {
			vp = (i << 11) + (char *)mbutl;
			printf("Neg cluster refcnt %d cluster #%d (%p)\n",
			    mclrefcnt[i], i, vp);
		}
	}

	if (!KVM_READ(nl[NL_MCLFREE].n_value, &mclfree, sizeof(mclfree)))
		errx(1, "kvm_read: %s", kvm_geterr(kd));
	printf("Scanning cluster free list...");
	fflush(stdout);
	i = 0;
	vp = mclfree;
	while (vp) {
		int clid = ((char *)vp - (char *)mbutl) >> 11;
		i++;
		if (clid < 0 || clid >= nmcl)
			errx(1, "Cluster pointer %p out of range!", vp);
		mclrefcnt[clid] = -1;
		if (!KVM_READ(vp, &vp, sizeof(vp)))
			errx(1, "kvm_read: %s", kvm_geterr(kd));
	}
	printf(" %d entries\n", i);

	minfo = calloc(nmbufs, sizeof(*minfo));
	if (!KVM_READ(nl[NL_MMBFREE].n_value, &mmbfree, sizeof(mmbfree)))
		errx(1, "kvm_read: %s", kvm_geterr(kd));
	printf("Scanning mbuf free list...");
	fflush(stdout);
	n = 0;
	vp = mmbfree;
	while (vp) {
		n++;
		i  = ((char *)vp - (char *)mbutl) / MSIZE;
		if (i < 0 || i >= nmbufs)
			errx(1, "Mbuf pointer %p out of range!", vp);
		minfo[i] |= MINFO_FREEMBUF;
		if (!KVM_READ(vp, &vp, sizeof(vp)))
			errx(1, "kvm_read: %s", kvm_geterr(kd));
	}
	printf(" %d entries\n", n);

	mprevi = calloc(nmbufs, sizeof(*mprevi));
	mbuf = malloc(nmbufs * sizeof(*mbuf));
	for (i = 0; i < nmbufs; i++) {
		mprevi[i] = -1;
	}
	for (i = 0; i < nmbufs; i++) {
		if (mclrefcnt[i/(MCLBYTES/MSIZE)])
			continue;


		if (!KVM_READ(&mbutl[i], &mbuf[i], sizeof(mbuf[i])))
			errx(1, "kvm_read mbutl[%d] @ %p: %s", i, &mbutl[i],
			    kvm_geterr(kd));

		if (mbuf[i].m_next) {
			int nexti;
			nexti = mbuf[i].m_next - mbutl;
			if (nexti < 0 || nexti > nmbufs) {
				printf("Weird nexti value %d for mbuf %p [%p]\n",
				    nexti, &mbutl[i], mbuf[i].m_next);
				continue;
			}
			mprevi[nexti] = i;
		}
		if (!(minfo[i] & MINFO_FREEMBUF))
			minfo[i] |= MINFO_VALIDMBUF;
	}
	printf("Mbuf chains:\n");
	for (index = 0; index < nmbufs; index++) {
		if (!(minfo[index] & MINFO_VALIDMBUF))
			continue;
		if (minfo[index] & (MINFO_SEEN | MINFO_FREEMBUF))
			continue;
		i = index;

		while (mprevi[i] != -1)
			i = mprevi[i];
		printf("Chain %p%s", &mbutl[i], dumpflag ? ": " : "\n");
		for (;;) {
			minfo[i] |= MINFO_SEEN;
			if (dumpflag)
				dump_mbuf(i, &mbuf[i], &mbutl[i]);
			else
				do_mbuf(i, &mbuf[i], &mbutl[i]);
			if (!mbuf[i].m_next)
				break;
			n = mbuf[i].m_next - mbutl;
			if (n < 0 || n > nmbufs) {
				printf("Weird next value %d for mbuf %p [%p]\n",
				    n, &mbutl[i], mbuf[i].m_next);
				break;
			}
			i = n;
		}
		if (dumpflag)
			printf("\n");
	}
	if (!dumpflag) {
		printf("\n\nMbufs in order\n");
		for (i = 0; i < nmbufs; i++) {
			if (mclrefcnt[i/(MCLBYTES/MSIZE)])
				continue;
			do_mbuf(i, &mbuf[i], &mbutl[i]);
		}
	}

	exit(0);
}

void do_mbuf(int i, struct mbuf *m, struct mbuf *m_kern) {
	struct mbuf *prevm = (mprevi[i] == -1) ? NULL : &mbutl[mprevi[i]];
	printf("\t%p len %d flags 0x%03x type %d next %p prev %p\n", m_kern,
	    m->m_len, m->m_flags, m->m_type, m->m_next, prevm);
}
	
void dump_mbuf(int i, struct mbuf *m, struct mbuf *m_kern) {
	unsigned char cl_data[MCLBYTES];
	unsigned char *cp;
	int n;

	if (m->m_data < (char *)&mbutl[i] ||
	    m->m_data + m->m_len > (char *)&mbutl[i+1]) {
		if (m->m_len > MCLBYTES) {
			printf(" Eeek, too long");
			return;
		}
		if (!KVM_READ(m->m_data, cl_data, m->m_len))
			errx(1, "dump_mbuf: kvm_read: %s", kvm_geterr(kd));
		cp = cl_data;
	} else
		cp = m->m_data - (char *)mbutl + (char *)mbuf;

	if (dumpflag == 1)
		for (n = 0; n < m->m_len; n++)
			printf(" %02x", cp[n]);
	else if (dumpflag == 2)
		fwrite(cp, 1, m->m_len, stdout);
}
	

void
usage()
{
	(void)fprintf(stderr,
 "usage: minfo [-N system] [-M core] [file ...]\n");
	exit(1);
}
